#!/usr/bin/env python
import sys
import asyncio
import argparse
from pathlib import Path

from azlassets import __version__, config, protobuf, versioncontrol, updater, repair, downloader
from azlassets.classes import Client


async def execute(args):
	# load config data from files
	userconfig = config.load_user_config()
	clientconfig = config.load_client_config(args.client)

	CLIENT_ASSET_DIR = Path(userconfig.asset_directory, args.client.name)
	CLIENT_ASSET_DIR.mkdir(parents=True, exist_ok=True)

	if args.check_integrity:
		update_assets = await repair.repair(clientconfig.cdnurl, userconfig, CLIENT_ASSET_DIR)

	if args.force_refresh and not args.repair:
		print("All asset types will be checked for different hashes.")

	version_response = protobuf.get_version_response(clientconfig.gateip, clientconfig.gateport)
	if not version_response:
		print("Server did not return a response to version request.")
		sys.exit(1)

	version_string = version_response.pb.version
	versionlist = [versioncontrol.parse_version_string(v) for v in version_string if v.startswith("$")]

	async with downloader.AzurlaneAsyncDownloader(clientconfig.cdnurl, useragent=userconfig.useragent) as downloader_session:
		for vresult in versionlist:
			if args.repair:
				update_assets = await repair.repair_hashfile(vresult, downloader_session, userconfig, CLIENT_ASSET_DIR)
			else:
				update_assets = await updater.update(vresult, downloader_session, userconfig, CLIENT_ASSET_DIR, args.force_refresh)

			if update_assets:
				versioncontrol.save_difflog2(vresult, update_assets, CLIENT_ASSET_DIR)


def main():
	print(f"Running Azurlane file downloader v{__version__}.")

	# setup argument parser
	parser = argparse.ArgumentParser()
	parser.add_argument("client", type=str, choices=Client.__members__,
		help="client to update")
	parser.add_argument("--force-refresh", type=bool, default=False, action=argparse.BooleanOptionalAction,
		help="compares asset hashes even when the version file is up to date")
	parser.add_argument("--repair", type=bool, default=False, action=argparse.BooleanOptionalAction,
		help="downloads missing files if the update process failed partially")
	parser.add_argument("--check-integrity", type=bool, default=False, action=argparse.BooleanOptionalAction,
		help="checks if all files are correct using the local hash file")
	args = parser.parse_args()

	args.client = Client[args.client]
	asyncio.run(execute(args))

if __name__ == "__main__":
	main()



import importlib
import socket


PROTOBUFS = {}
def import_pb(pb: int, addname: str = ""):
	module = importlib.import_module(f".p{pb}{addname}_pb_pb2", "azlassets.proto")
	PROTOBUFS[pb] = module

def import_pb_with_retry(pb: int):
	try:
		import_pb(pb)
	except ModuleNotFoundError:
		import_pb(pb, "min")

class InvalidHeaderError(Exception): pass

class BasicCommand:
	def __init__(self, command_id: int, index: int=0):
		pbpackage = command_id//1000
		if pbpackage not in PROTOBUFS:
			try:
				import_pb_with_retry(pbpackage)
			except ModuleNotFoundError:
				print(f'Command Package {pbpackage} does not exist (From cmd={command_id}).')

		pbpak = PROTOBUFS[pbpackage]
		if hasattr(pbpak, 'cs_'+str(command_id)):
			self.command_id = command_id
			self.index = index
			self.pb = getattr(pbpak, 'cs_'+str(command_id))()
		elif hasattr(pbpak, 'sc_'+str(command_id)):
			self.command_id = command_id
			self.index = index
			self.pb = getattr(pbpak, 'sc_'+str(command_id))()
		else:
			print(f'Command {command_id} is not registered.')


ADV_HEADER_LEN = 7
HEADER_LEN = 2
HEADER_NOID_LEN = ADV_HEADER_LEN - HEADER_LEN

def serialize_pb(basic_pb_command):
	payload = bytearray(basic_pb_command.pb.SerializeToString())

	header_bytes = ((len(payload) or 1) + HEADER_NOID_LEN).to_bytes(2, byteorder='big')
	command_id_bytes = basic_pb_command.command_id.to_bytes(2, byteorder='big')
	index_bytes = basic_pb_command.index.to_bytes(2, byteorder='big')

	command_bytes = bytearray(ADV_HEADER_LEN)
	command_bytes[0] = header_bytes[0]
	command_bytes[1] = header_bytes[1]
	command_bytes[3] = command_id_bytes[0]
	command_bytes[4] = command_id_bytes[1]
	command_bytes[5] = index_bytes[0]
	command_bytes[6] = index_bytes[1]
	command_bytes.extend(payload)
	return command_bytes


def deserialize_header(header_bytes):
	if not header_bytes[2] == 0:
		raise InvalidHeaderError('Received invalid header.')
	payload_size = (header_bytes[0] << 8 | header_bytes[1]) - HEADER_NOID_LEN
	command_id = header_bytes[3] << 8 | header_bytes[4]
	index = header_bytes[5] << 8 | header_bytes[6]
	return payload_size, command_id, index

def deserialize_pb(command_id, index, payload_bytes):
	basiccmd = BasicCommand(command_id)
	if hasattr(basiccmd, 'pb'):
		basiccmd.pb.ParseFromString(payload_bytes)
		print(f'Successfully received cmd={command_id} with idx={index}.')
		return basiccmd


# PROTOBUF HELPER METHODS
class ConnectionSocket(socket.socket):
	def __init__(self, blocking=True):
		super().__init__(socket.AF_INET, socket.SOCK_STREAM)
		self.setblocking(blocking)

	def disconnect(self):
		self.close()

	def send_command(self, commandid, index, **kwargs):
		print(f"Sending Command {commandid}.")
		command = BasicCommand(commandid, index)
		for k, v in kwargs.items():
			setattr(command.pb, k, v)
		self.send(serialize_pb(command))

	def recv_command(self):
		header = self.recv_bytes(ADV_HEADER_LEN)
		payload_size, command_id, index = deserialize_header(header)
		payload = self.recv_bytes(payload_size)
		return deserialize_pb(command_id, index, payload)

	def recv_bytes(self, bytesize:int):
		data = b''
		while len(data) < bytesize:
			data += self.recv(bytesize - len(data))
		return data

def get_version_response(gateip, gateport):
	with ConnectionSocket() as socket:
		socket.connect((gateip, gateport))
		socket.send_command(
			10800,
			0,
			state=21,
			platform='0'
		)
		result = socket.recv_command()
	return result



import json
from pathlib import Path
from typing import Generator, Iterable

from .classes import DownloadType, HashRow, UpdateResult, VersionType, VersionResult


def parse_version_string(rawstring: str) -> VersionResult:
	"""
	Tries to parse the raw version string as returned by the game server into a VersionResult.

	Raises NotImplementedError if the versiontype does not exist.
	"""
	parts = rawstring.split('$')[1:]
	versionname = parts[0]
	versiontype = VersionType.from_hashname(versionname)
	if not versiontype:
		raise NotImplementedError(f'Unknown versionname {versionname}.')

	if versiontype == VersionType.AZL:
		version = '.'.join(parts[1:-1])
		return VersionResult(version, parts[-1], rawstring, versiontype)
	return VersionResult(parts[1], parts[2], rawstring, versiontype)

def compare_version_string(version_new: str, version_old: str | None) -> bool:
	"""
	Returns `True` if `version_new` is newer than `version_old`.
	"""
	if not version_old:
		return True
	new_to_int = [int(v) for v in version_new.split(".")]
	old_to_int = [int(v) for v in version_old.split(".")]
	return new_to_int > old_to_int


def load_version_string(version_type: VersionType, relative_parent_dir: Path) -> str | None:
	fpath = Path(relative_parent_dir, version_type.version_filename)
	if fpath.exists():
		with open(fpath, 'r', encoding='utf8') as f:
			return f.read()

def save_version_string(version_type: VersionType, relative_parent_dir: Path, content: str):
	with open(Path(relative_parent_dir, version_type.version_filename), 'w', encoding='utf8') as f:
		f.write(content)

def save_version_string2(version_result: VersionResult, relative_parent_dir: Path):
	save_version_string(version_result.version_type, relative_parent_dir, version_result.version)


def iterate_hash_lines(hashes: str) -> Generator[list[str], None, None]:
	for assetinfo in hashes.splitlines():
		if assetinfo == '': continue
		yield assetinfo.split(',')

def parse_hash_rows(hashes: str) -> Generator[HashRow, None, None]:
	for path, size, md5hash in iterate_hash_lines(hashes):
		yield HashRow(path, int(size), md5hash)

def load_hash_file(version_type: VersionType, relative_parent_dir: Path) -> Generator[HashRow, None, None] | None:
	fpath = Path(relative_parent_dir, version_type.hashes_filename)
	if fpath.exists():
		with open(fpath, 'r', encoding='utf8') as f:
			return parse_hash_rows(f.read())

def save_hash_file(version_type: VersionType, relative_parent_dir: Path, hashrows: Iterable[HashRow]):
	rowstrings = [f"{row.filepath},{row.size},{row.md5hash}" for row in hashrows if row]
	content = '\n'.join(rowstrings)
	with open(Path(relative_parent_dir, version_type.hashes_filename), 'w', encoding="utf8") as f:
		f.write(content)

def update_version_data(version_type: VersionType, relative_parent_dir: Path, version_string: str, hashrows: Iterable[HashRow]):
	save_version_string(version_type, relative_parent_dir, version_string)
	save_hash_file(version_type, relative_parent_dir, hashrows)

def update_version_data2(version_result: VersionResult, relative_parent_dir: Path, hashrows: Iterable[HashRow]):
	save_version_string2(version_result, relative_parent_dir)
	save_hash_file(version_result.version_type, relative_parent_dir, hashrows)


def get_latest_versionstring(version_type: VersionType, relative_parent_dir: Path) -> str | None:
	version_diffdir = Path(relative_parent_dir, "difflog", version_type.name.lower())

	legacy_rename_latest_difflog(version_diffdir)

	latest_versionfile = Path(version_diffdir, "latest")
	if latest_versionfile.exists():
		with open(latest_versionfile, "r", encoding="utf8") as f:
			return f.read()

def save_difflog(version_type: VersionType, version_string: str, update_results: list[UpdateResult], relative_parent_dir: Path):
	filtered_update_results = list(filter(lambda r: r.download_type != DownloadType.NoChange, update_results))
	if not filtered_update_results:
		return

	version_diffdir = Path(relative_parent_dir, "difflog", version_type.name.lower())
	version_diffdir.mkdir(parents=True, exist_ok=True)
	legacy_rename_latest_difflog(version_diffdir)

	data = {
		"version": version_string,
		"major": False,
		"success_files": {res.path.inner: res.compare_result.compare_type.name for res in filter(lambda r: r.download_type in [DownloadType.Success, DownloadType.Removed], filtered_update_results)},
		"failed_files": {res.path.inner: res.compare_result.compare_type.name for res in filter(lambda r: r.download_type == DownloadType.Failed, filtered_update_results)},
	}

	difflog_filepath = Path(version_diffdir, version_string+".json")
	with open(difflog_filepath, "w", encoding="utf8") as f:
		json.dump(data, f)

	latest_filepath = Path(version_diffdir, "latest")
	with open(latest_filepath, "w", encoding="utf8") as f:
		f.write(version_string)

def legacy_rename_latest_difflog(version_diffdir: Path):
	"""
	Rename the `latest.json` difflog file for users that used the tool while that file was being created.
	"""
	latest_difflog = Path(version_diffdir, "latest.json")
	if latest_difflog.exists():
		with open(latest_difflog, "r", encoding="utf8") as f:
			latest_data = json.load(f)
		latest_version = latest_data["version"]
		latest_difflog.rename(latest_difflog.with_stem(latest_version))

		latest_filepath = Path(version_diffdir, "latest")
		with open(latest_filepath, "w", encoding="utf8") as f:
			f.write(latest_version)


def save_difflog2(version_result: VersionResult, update_results: list[UpdateResult], relative_parent_dir: Path):
	save_difflog(version_result.version_type, version_result.version, update_results, relative_parent_dir)



import asyncio
from pathlib import Path
from typing import Iterable
from collections import defaultdict

from . import downloader, versioncontrol
from .classes import *


def remove_asset(filepath: Path):
	if filepath.exists():
		filepath.unlink()
	else:
		print(f"WARN: Tried to remove non-existant asset at {filepath}")


downloader_semaphore = asyncio.Semaphore(6)

async def handle_asset_download(downloader_session: downloader.AzurlaneAsyncDownloader, assetbasepath: Path, result: CompareResult,
								progressbar: ProgressBar | None = None) -> UpdateResult:
	newhash = result.new_hash
	assert newhash is not None

	assetpath = BundlePath.construct(assetbasepath, newhash.filepath)
	async with downloader_semaphore:  # prevent queueing into connection pool, since wait time in pool counts towards timeout
		download_success = await downloader_session.download_asset(newhash.md5hash, assetpath.full, newhash.size)

	if progressbar:
		progressbar.update()
	return UpdateResult(result, DownloadType.Success if download_success else DownloadType.Failed, assetpath)


async def update_assets(downloader_session: downloader.AzurlaneAsyncDownloader, comparison_results: dict[CompareType, list[CompareResult]],
						client_directory: Path, allow_deletion: bool = True) -> list[UpdateResult]:
	assetbasepath = client_directory / "AssetBundles"
	update_results = [UpdateResult(r, DownloadType.NoChange, BundlePath.construct(assetbasepath, r.new_hash.filepath)) for r in comparison_results[CompareType.Unchanged]]

	# handle all new or changed files
	update_files = comparison_results[CompareType.New] + comparison_results[CompareType.Changed]
	if len(update_files) > 0:
		progressbar = ProgressBar(len(update_files), "Download Progress", details_unit="files")
		tasks = [handle_asset_download(downloader_session, assetbasepath, result, progressbar) for result in update_files]
		update_results += await asyncio.gather(*tasks)

	# handle all deleted files
	deleted_files = comparison_results[CompareType.Deleted]
	if len(deleted_files) > 0:
		if allow_deletion:
			progressbar = ProgressBar(len(deleted_files), "Deletion Progress", details_unit="files")
			for result in deleted_files:
				assetpath = BundlePath.construct(assetbasepath, result.current_hash.filepath)
				remove_asset(assetpath.full)
				update_results.append(UpdateResult(result, DownloadType.Removed, assetpath))
				progressbar.update()
		else:
			for result in deleted_files:
				assetpath = BundlePath.construct(assetbasepath, result.current_hash.filepath)
				update_results.append(UpdateResult(result, DownloadType.ForDeletionNoChange, assetpath))

	return update_results

async def download_and_parse_hashes(version_result: VersionResult, downloader_session: downloader.AzurlaneAsyncDownloader, userconfig: UserConfig) -> list[HashRow] | None:
	hashes = await downloader_session.download_hashes(version_result)
	if not hashes:
		print(f"Server returned empty hashfile for {version_result.version_type.name}, skipping.")
		return

	# hash filter function
	def _filter(row: HashRow):
		for path in userconfig.download_filter:
			if row.filepath.startswith(path):
				if not userconfig.download_isblacklist:
					return True
		return userconfig.download_isblacklist

	return list(filter(_filter, versioncontrol.parse_hash_rows(hashes)))

def compare_hashes(oldhashes: Iterable[HashRow], newhashes: Iterable[HashRow]) -> dict[CompareType, list[CompareResult]]:
	results = {row.filepath: CompareResult(None, row, CompareType.New) for row in newhashes}
	for hashrow in oldhashes:
		res = results.get(hashrow.filepath)
		if res is None:
			results[hashrow.filepath] = CompareResult(hashrow, None, CompareType.Deleted)
		elif hashrow == res.new_hash:
			res.current_hash = hashrow
			res.compare_type = CompareType.Unchanged
		else: # file has changed
			res.current_hash = hashrow
			res.compare_type = CompareType.Changed

	sorted_results = defaultdict(list)
	for r in results.values():
		sorted_results[r.compare_type].append(r)
	return sorted_results

def filter_hashes(update_results: list[UpdateResult]) -> list[HashRow]:
	hashes_updated = []
	for update_result in update_results:
		if update_result.download_type in [DownloadType.Success, DownloadType.NoChange]:
			hashrow = update_result.compare_result.new_hash
		elif update_result.download_type == DownloadType.ForDeletionNoChange:
			hashrow = update_result.compare_result.current_hash
		elif update_result.download_type == DownloadType.Failed:
			hashrow = update_result.compare_result.current_hash
			if not hashrow:
				continue
		else:
			continue

		# some error checking, although it should not be needed anymore
		if hashrow:
			hashes_updated.append(hashrow)
		else:
			print("WARN: Empty hashrow detected while it should not have been empty. Debug info below.")
			print(update_result)
	return hashes_updated

async def _update_from_hashes(version_result: VersionResult, downloader_session: downloader.AzurlaneAsyncDownloader, client_directory: Path,
						oldhashes: Iterable[HashRow], newhashes: Iterable[HashRow], allow_deletion: bool = True) -> list[UpdateResult]:
	comparison_results = compare_hashes(oldhashes, newhashes)
	update_results = await update_assets(downloader_session, comparison_results, client_directory, allow_deletion)
	hashes_updated = filter_hashes(update_results)
	versioncontrol.update_version_data2(version_result, client_directory, hashes_updated)
	return update_results

async def _update(version_result: VersionResult, downloader_session: downloader.AzurlaneAsyncDownloader, userconfig: UserConfig, client_directory: Path) -> list[UpdateResult] | None:
	newhashes = await download_and_parse_hashes(version_result, downloader_session, userconfig)
	if newhashes:
		oldhashes = versioncontrol.load_hash_file(version_result.version_type, client_directory)
		return await _update_from_hashes(version_result, downloader_session, client_directory, oldhashes or [], newhashes)

async def update(version_result: VersionResult, downloader_session: downloader.AzurlaneAsyncDownloader, userconfig: UserConfig, client_directory: Path, force_refresh: bool) -> list[UpdateResult] | None:
	oldversion = versioncontrol.load_version_string(version_result.version_type, client_directory)
	if versioncontrol.compare_version_string(version_result.version, oldversion):
		print(f"{version_result.version_type.name}: Current version {oldversion} is older than latest version {version_result.version}.")
		return await _update(version_result, downloader_session, userconfig, client_directory)
	else:
		print(f"{version_result.version_type.name}: Current version {oldversion} is latest.")
		if force_refresh:
			return await _update(version_result, downloader_session, userconfig, client_directory)



import hashlib
import itertools
import aiofile
import asyncio
from pathlib import Path

from . import downloader, updater, versioncontrol
from .classes import *


async def execute_coro_with_progressbar(coro, progressbar: ProgressBar):
	r = await coro
	progressbar.update()
	return r


semaphore_concurrent_files = asyncio.Semaphore(5)

async def calc_md5hash(filepath: Path, chunk_size: int = 65536) -> str:
	md5 = hashlib.md5()
	async with semaphore_concurrent_files:
		async with aiofile.async_open(filepath, "rb") as f:
			async for chunk in f.iter_chunked(chunk_size):
				if chunk:
					md5.update(chunk)
	return md5.hexdigest()

async def get_filedata(filepath: Path) -> tuple[str, int]:
	if filepath.exists():
		current_md5 = await calc_md5hash(filepath)
		current_size = filepath.stat().st_size
		return current_md5, current_size
	else:
		return "", 0

async def hashrow_from_file(assetbasepath: Path, filepath: Path) -> HashRow:
	current_md5, current_size = await get_filedata(filepath)
	clean_filepath = str(filepath.relative_to(assetbasepath)).replace("\\", "/")
	return HashRow(clean_filepath, current_size, current_md5)

async def hashrow_from_relative_file(assetbasepath: Path, relative_filepath: Path) -> HashRow:
	current_md5, current_size = await get_filedata(assetbasepath / relative_filepath)
	return HashRow(relative_filepath, current_size, current_md5)

async def hashrows_from_files(client_directory: Path) -> list[HashRow]:
	assetbasepath = client_directory / "AssetBundles"
	progressbar = ProgressBar(0, "File Progress", details_unit="files", print_on_init=False)
	print("Loading list of all files... ", end="")
	tasks = [execute_coro_with_progressbar(hashrow_from_file(assetbasepath, fp), progressbar) for fp in assetbasepath.rglob("*") if not fp.is_dir()]
	progressbar.total = len(tasks)
	print("Done")
	print("Checking all files...")
	return await asyncio.gather(*tasks)

async def repair(cdnurl: str, userconfig: UserConfig, client_directory: Path) -> list[UpdateResult]:
	current_hashes = await hashrows_from_files(client_directory)
	expected_hashes = itertools.chain(*filter(lambda x: x is not None, [versioncontrol.load_hash_file(vtype, client_directory) for vtype in VersionType]))
	comparison_results = updater.compare_hashes(current_hashes, expected_hashes)
	async with downloader.AzurlaneAsyncDownloader(cdnurl, useragent=userconfig.useragent) as downloader_session:
		update_results = await updater.update_assets(downloader_session, comparison_results, client_directory)
	return update_results

async def repair_hashfile(version_result: VersionResult, cdnurl: str, userconfig: UserConfig, client_directory: Path) -> list[UpdateResult]:
	# read hashes that are stored in the local hash file
	localhashes = versioncontrol.load_hash_file(version_result.version_type, client_directory)

	async with downloader.AzurlaneAsyncDownloader(cdnurl, useragent=userconfig.useragent) as downloader_session:
		# load newest hashes from the game server
		serverhashes = await updater.download_and_parse_hashes(version_result, downloader_session, userconfig) or []
		assetbasepath = client_directory / "AssetBundles"

		# parse hashes from all files stored on disk, but only check files that are expected based on the new hashes
		# this skips deletion on unneeded files
		print("Generating hashes for all files on disk...")
		progressbar = ProgressBar(len(serverhashes), "File Progress", details_unit="files")
		diskhashes_tasks = [execute_coro_with_progressbar(hashrow_from_relative_file(assetbasepath, hrow.filepath), progressbar) for hrow in serverhashes]
		diskhashes = await asyncio.gather(*diskhashes_tasks)

		# compare localhashes to diskhashes to determine which files have already been successfully downloaded
		compare_results_disk = updater.compare_hashes(localhashes, diskhashes)
		update_results_disk = {comp_result.new_hash.filepath: UpdateResult(comp_result, DownloadType.Success, comp_result.new_hash.filepath) for comp_result in compare_results_disk[CompareType.Changed]}
		update_results_disk.update({comp_result.new_hash.filepath: UpdateResult(comp_result, DownloadType.Success, comp_result.new_hash.filepath) for comp_result in compare_results_disk[CompareType.New]})
		update_results_disk.update({comp_result.current_hash.filepath: UpdateResult(comp_result, DownloadType.Removed, comp_result.current_hash.filepath) for comp_result in compare_results_disk[CompareType.Deleted]})

		# download remaining files
		update_results_server = await updater._update_from_hashes(version_result, downloader_session, client_directory, diskhashes, serverhashes, allow_deletion=False)

		# add old update results to new update results list
		update_results = []
		for upres_server in update_results_server:
			update_result = upres_server
			# try to retrieve from old list only if there was no further change to the file
			if upres_server.download_type == DownloadType.NoChange:
				if upres_disk := update_results_disk.get(upres_server.path):
					update_result = upres_disk
			update_results.append(update_result)

	return update_results



import aiohttp
import aiofile
import traceback
from pathlib import Path

from .classes import VersionResult


class AzurlaneAsyncDownloader(aiohttp.ClientSession):
	def __init__(self, cdn_url: str, useragent: str):
		base_url = f"{cdn_url}/android/"
		limited_tcp_connector = aiohttp.TCPConnector(limit_per_host=6)
		super().__init__(base_url=base_url, headers={"user-agent": useragent}, connector=limited_tcp_connector)

	async def get_hashes(self, versionhash: str) -> aiohttp.ClientResponse:
		return await self.get(f"hash/{versionhash}")

	async def get_asset(self, filehash: str) -> aiohttp.ClientResponse:
		return await self.get(f"resource/{filehash}")

	async def download_hashes(self, version_result: VersionResult) -> str | None:
		try:
			async with await self.get_hashes(version_result.rawstring) as response:
				response: aiohttp.ClientResponse
				response.raise_for_status() # raises error on bad HTTP status

				hashes = await response.text()
				return hashes

		except Exception as e:
			print(f"ERROR: An unexpected error occured while downloading '{version_result.version_type.name}' hashfile.")
			traceback.print_exception(type(e), e, e.__traceback__)
			return

	async def download_asset(self, filehash: str, save_destination: Path, expected_file_size: int) -> bool:
		"""
		Downloads the requested file using the session and saves it to 'save_destination' on disk.

		Returns `True` if the operation was successful, otherwise `False`.
		"""
		try:
			async with await self.get_asset(filehash) as response:
				response: aiohttp.ClientResponse
				response.raise_for_status() # raises error on bad HTTP status

				# reject response if response size doesn't match expected size
				response_size = response.content_length
				if expected_file_size != response_size:
					print(f"ERROR: Received asset '{filehash}' with target '{save_destination}' has wrong size ({response_size}/{expected_file_size}).")
					return False

				save_destination.parent.mkdir(parents=True, exist_ok=True)
				async with aiofile.async_open(save_destination, "wb") as file:
					async for chunk in response.content.iter_chunked(1024*16): # no idea what chuck size is best
						await file.write(chunk)

			return True
		except Exception as e:
			print(f"ERROR: An unexpected error occured while downloading '{filehash}' to '{save_destination}'.")
			traceback.print_exception(type(e), e, e.__traceback__)
			return False



from enum import Enum
from dataclasses import dataclass
from pathlib import Path
from typing import Self


CompareType = Enum('CompareType', 'New Changed Unchanged Deleted')
DownloadType = Enum('DownloadType', 'NoChange Removed Success Failed ForDeletionNoChange')

class VersionType(Enum):
	__hash2member_map__: dict[str, Self] = {}
	hashname: str
	"""Hash name used on the version result returned by the game server."""
	suffix: str
	"""Suffix used on version and hash files."""

	AZL			= (1,	"azhash",		"")
	CV			= (2,	"cvhash",		"cv")
	L2D			= (3,	"l2dhash",		"live2d")
	PIC			= (4,	"pichash",		"pic")
	BGM			= (5,	"bgmhash",		"bgm")
	CIPHER		= (6,	"cipherhash",	"cipher")
	MANGA		= (7,	"mangahash",	"manga")
	PAINTING	= (8,	"paintinghash",	"painting")
	DORM		= (9,	"dormhash",		"dorm")
	MAP			= (10,	"maphash",		"map")


	def __init__(self, _, hashname, suffix) -> None:
		# add attributes to enum objects
		self.hashname = hashname
		self.suffix = suffix
		# add enum objects to member maps
		self.__hash2member_map__[hashname] = self

	def __str__(self) -> str:
		return self.name.lower()

	@property
	def version_filename(self) -> str:
		"""
		Full version filename using the suffix.
		"""
		suffix = self.suffix
		if suffix: suffix = "-"+suffix
		return f"version{suffix}.txt"

	@property
	def hashes_filename(self) -> str:
		"""
		Full hashes filename using the suffix.
		"""
		suffix = self.suffix
		if suffix: suffix = "-"+suffix
		return f"hashes{suffix}.csv"

	@classmethod
	def from_hashname(cls, hashname: str) -> Self | None:
		"""
		Returns a VersionType member with matching *hashname* if match exists, otherwise None.
		"""
		return cls.__hash2member_map__.get(hashname)


class AbstractClient(Enum):
	active: bool
	locale_code: str
	package_name: str

	def __new__(cls, value, active, locale, package_name):
		# this should be done differently, but i am too lazy to do that now
		# TODO: change it
		if not hasattr(cls, "package_names"):
			cls.package_names = {}

		obj = object.__new__(cls)
		obj._value_ = value
		obj.active = active
		obj.locale_code = locale
		obj.package_name = package_name
		cls.package_names[package_name] = obj
		return obj

	@classmethod
	def from_package_name(cls, package_name) -> Self | None:
		return cls.package_names.get(package_name)


class Client(AbstractClient):
	EN = (1, True, 'en-US', 'com.YoStarEN.AzurLane')
	JP = (2, True, 'ja-JP', 'com.YoStarJP.AzurLane')
	CN = (3, True, 'zh-CN', '')
	KR = (4, True, 'ko-KR', 'kr.txwy.and.blhx')
	TW = (5, True, 'zh-TW', 'com.hkmanjuu.azurlane.gp')


@dataclass
class HashRow:
	filepath: str
	size: int
	md5hash: str

@dataclass
class CompareResult:
	current_hash: HashRow | None
	new_hash: HashRow | None
	compare_type: CompareType

@dataclass
class VersionResult:
	version: str
	vhash: str
	rawstring: str
	version_type: VersionType

@dataclass
class BundlePath:
	full: Path
	inner: str

	@staticmethod
	def construct(parentdir: Path, inner: Path | str) -> "BundlePath":
		fullpath = Path(parentdir, inner)
		return BundlePath(fullpath, str(inner))

	def __hash__(self):
		return hash(self.inner)

@dataclass
class UpdateResult:
	compare_result: CompareResult
	download_type: DownloadType
	path: BundlePath

@dataclass
class UserConfig:
	useragent: str
	download_isblacklist: bool
	download_filter: list
	extract_isblacklist: bool
	extract_filter: list
	asset_directory: Path
	extract_directory: Path

@dataclass
class ClientConfig:
	gateip: str
	gateport: int
	cdnurl: str


# stolen from https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
def printProgressBar(iteration, total, prefix = '', suffix = 'Complete', decimals = 1, length = 50, fill = '█', printEnd = "\r", details_unit = None):
	"""
	Call in a loop to create terminal progress bar
	@params:
		iteration	- Required  : current iteration (Int)
		total		- Required  : total iterations (Int)
		prefix		- Optional  : prefix string (Str)
		suffix		- Optional  : suffix string (Str)
		decimals	- Optional  : positive number of decimals in percent complete (Int)
		length		- Optional  : character length of bar (Int)
		fill		- Optional  : bar fill character (Str)
		printEnd	- Optional  : end character (e.g. "\r", "\r\n") (Str)
	"""
	percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
	filledLength = int(length * iteration // total)
	progress = fill * filledLength + '-' * (length - filledLength)
	details = f" [{iteration}/{total} {details_unit}]" if details_unit else ""
	print(f'\r{prefix} |{progress}| {percent}% {suffix}{details}', end = printEnd)
	# Print New Line on Complete
	if iteration == total:
		print()

# simplified progress bar class with only the useful stuff i use
class ProgressBar():
	def __init__(self, total: int, prefix: str, suffix: str = "Complete", iterstart: int = 0, details_unit: str | None = None, print_on_init: bool = True):
		self.iteration = iterstart
		self.total = total
		self.prefix = prefix
		self.suffix = suffix
		self.details_unit = details_unit
		if print_on_init:
			printProgressBar(iterstart, total, prefix, suffix, details_unit=details_unit)

	def update(self, iteration: int | None = None):
		if iteration:
			self.iteration = iteration
		else:
			self.iteration += 1
		printProgressBar(self.iteration, self.total, self.prefix, self.suffix, details_unit=self.details_unit)


__all__ = [
	"CompareType",
	"DownloadType",
	"VersionType",
	"Client",
	"HashRow",
	"CompareResult",
	"VersionResult",
	"BundlePath",
	"UpdateResult",
	"UserConfig",
	"ClientConfig",
	"ProgressBar",
]



#!/usr/bin/env python
import io, re, sys, json, shutil, hashlib, itertools
from argparse import ArgumentParser, BooleanOptionalAction
from zipfile import ZipFile
from pathlib import Path
from collections import defaultdict

from azlassets import __version__, versioncontrol, updater, config
from azlassets.classes import BundlePath, Client, CompareType, DownloadType, UpdateResult, VersionType, ProgressBar


def calc_md5hash(data: bytes) -> str:
	md5 = hashlib.md5()
	md5.update(data)
	return md5.hexdigest()


def unpack(zipfile: ZipFile, client: Client, allow_older_version: bool = False):
	# load config data from files
	userconfig = config.load_user_config()
	CLIENT_ASSET_DIR = Path(userconfig.asset_directory, client.name)
	CLIENT_ASSET_DIR.mkdir(parents=True, exist_ok=True)

	# create {filename: filesize} dict for later recovery of missed files
	file_info_list = {f.filename: f.file_size for f in zipfile.filelist if not f.is_dir()}

	print('Unpacking archive...')
	for versiontype in VersionType:
		# make sure the version file exists
		if not 'assets/'+versiontype.version_filename in zipfile.namelist():
			print(f'{versiontype.name}: The file {versiontype.version_filename} could not be found in the archive. Has the archive been modified?')
			continue

		# read version string from obb
		with zipfile.open('assets/'+versiontype.version_filename, 'r') as zf:
			obbversion = zf.read().decode('utf8')

		# if the obbversion is older, don't extract data from obb
		currentversion = versioncontrol.load_version_string(versiontype, CLIENT_ASSET_DIR)
		if not versioncontrol.compare_version_string(obbversion, currentversion):
			print(f'{versiontype.name}: Current version {currentversion} is same or newer than obb version {obbversion}.')
			continue

		# read hash files from obb and current file and compare them
		with zipfile.open('assets/'+versiontype.hashes_filename, 'r') as hashfile:
			obbhashes = versioncontrol.parse_hash_rows(hashfile.read().decode('utf8'))
		currenthashes = versioncontrol.load_hash_file(versiontype, CLIENT_ASSET_DIR)
		comparison_results = updater.compare_hashes(currenthashes or [], obbhashes)

		# extract and delete files
		assetbasepath = Path(CLIENT_ASSET_DIR, 'AssetBundles')
		update_files = list(itertools.chain(*[v for comp_type, v in comparison_results.items() if comp_type != CompareType.Unchanged]))
		update_results = [UpdateResult(r, DownloadType.NoChange, BundlePath.construct(assetbasepath, r.new_hash.filepath)) for r in comparison_results[CompareType.Unchanged]]

		fileamount = len(update_files)
		if fileamount > 0:
			files_not_found = []
			progressbar = ProgressBar(fileamount, f"Extracting '{versiontype.hashname}' Files", details_unit="files")
			for result in update_files:
				if result.compare_type in [CompareType.New, CompareType.Changed]:
					assetpath = BundlePath.construct(assetbasepath, result.new_hash.filepath)
					if pathresult := extract_asset(zipfile, assetpath.inner, assetpath.full):
						file_info_list.pop(pathresult)
						update_results.append(UpdateResult(result, DownloadType.Success if assetpath.full.exists() else DownloadType.Failed, assetpath))
					else:
						files_not_found.append((assetpath, result))
				elif result.compare_type == CompareType.Deleted:
					assetpath = BundlePath.construct(assetbasepath, result.current_hash.filepath)
					updater.remove_asset(assetpath.full)
					update_results.append(UpdateResult(result, DownloadType.Removed, assetpath))
				progressbar.update()

			# try to find remaining files using their md5hash
			if len(files_not_found) > 0:
				file_info_groupby_size = defaultdict(list)
				for k,v in file_info_list.items():
					file_info_groupby_size[v].append(k)

				progressbar = ProgressBar(len(files_not_found), "Retrieving failed files", details_unit="files")
				for assetpath,result in files_not_found:
					fileinfo = file_info_groupby_size.get(result.new_hash.size)
					if not fileinfo:
						continue

					for zipf_path in fileinfo:
						with zipfile.open(zipf_path, "r") as zf:
							zipf_data = zf.read()
							zipf_md5hash = calc_md5hash(zipf_data)
							if result.new_hash.md5hash == zipf_md5hash:
								with open(assetpath.full, "wb") as f:
									f.write(zipf_data)
									update_results.append(UpdateResult(result, DownloadType.Success if assetpath.full.exists() else DownloadType.Failed, assetpath))
									break
					#else:
					#	print( LOG ERROR MESSAGE HERE )
					progressbar.update()


		# update version string, hashes and difflog
		hashes_updated = updater.filter_hashes(update_results)
		versioncontrol.update_version_data(versiontype, CLIENT_ASSET_DIR, obbversion, hashes_updated)
		versioncontrol.save_difflog(versiontype, obbversion, update_results, CLIENT_ASSET_DIR)


def extract_asset(zipfile: ZipFile, filepath: str, target: Path):
	target.parent.mkdir(exist_ok=True, parents=True)

	if "." in Path(filepath).name:
		assetpath = "assets/AssetBundles/"+filepath
	else:
		assetpath = "assets/AssetBundles/"+filepath+".ys"

	try:
		with zipfile.open(assetpath, 'r') as zf, open(target, 'wb') as f:
			shutil.copyfileobj(zf, f)
			return assetpath
	except KeyError:
		pass


def extract_obb(path: Path, fallback_client: Client | None = None, allow_older_version: bool = False):
	for client in Client:
		if client.package_name and re.match(rf".*{client.package_name}\.obb", path.name):
			print(f'Determined client {client.name} from filename.')
			with ZipFile(path, 'r') as zipfile:
				unpack(zipfile, client, allow_older_version)
			break
	else:
		if fallback_client:
			print(f"Unpacking using provided client {fallback_client.name}.")
			with ZipFile(path, 'r') as zipfile:
				unpack(zipfile, fallback_client, allow_older_version)
		else:
			sys.exit(f'Filename "{path.name}" could not be associated with any known client.')

def extract_xapk(path: Path, allow_older_version: bool = False):
	with ZipFile(path, 'r') as xapk_archive:
		# read the manifest file for information about the client, obbs and the apk paths inside the archive
		with xapk_archive.open('manifest.json', 'r') as manifestfile:
			manifest = json.loads(manifestfile.read().decode('utf8'))

		# determine the client the xpak comes from
		if client := Client.from_package_name(manifest['package_name']):
			print(f'Determined client {client.name} from manifest.json file.')
			# find all obb archives and extract files from them
			for obb_expansion in manifest['expansions']:
				with xapk_archive.open(obb_expansion['file'], 'r') as mainobbfile:
					# need to load the full obb into memory, otherwise it has to constantly read the whole obb again and performance gets shit
					mainobb_filedata = io.BytesIO(mainobbfile.read())

					with ZipFile(mainobb_filedata, 'r') as main_obb:
						unpack(main_obb, client, allow_older_version)
		else:
			print("Could not determine client from xapk manifest.json.")


def extract(path: Path, fallback_client: Client | None = None, allow_older_version: bool = False):
	if not path.exists():
		sys.exit('This file does not exist.')

	if path.suffix == '.obb':
		print('File has .obb extension.')
		extract_obb(path, fallback_client, allow_older_version)
	elif path.suffix == '.apk':
		if fallback_client:
			apk_client = fallback_client
			print(f"File has .apk extension, using provided fallback client {apk_client.name}.")
		else:
			apk_client = Client.CN
			print(f"File has .apk extension and no fallback client has been provided, assuming {apk_client.name} client.")

		with ZipFile(path, 'r') as zipfile:
			unpack(zipfile, apk_client, allow_older_version)
	elif path.suffix == '.xapk':
		print('File has .xapk extension.')
		extract_xapk(path, allow_older_version)
	else:
		sys.exit(f'Unknown file extension "{path.suffix}".')


def main():
	print(f"Running Azurlane obb_apk_import v{__version__}.")

	parser = ArgumentParser()
	parser.add_argument('file', nargs=1, help='xapk/apk/obb file to extract')
	parser.add_argument('-c', '--client', help='fallback client if it cannot be determined automatically (obb/apk only)', choices=Client.__members__)
	parser.add_argument('--allow_old_version', help='when enabled, the old version check is not used', action=BooleanOptionalAction)
	args = parser.parse_args()
	extract(Path(args.file[0]), args.client, args.allow_old_version)

if __name__ == "__main__":
	main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: p10min_pb.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fp10min_pb.proto\x12\x0f\x61zlassets.proto\"+\n\x08\x63s_10800\x12\r\n\x05state\x18\x01 \x02(\r\x12\x10\n\x08platform\x18\x02 \x02(\t\"\xce\x01\n\x08sc_10801\x12\x12\n\ngateway_ip\x18\x01 \x02(\t\x12\x14\n\x0cgateway_port\x18\x02 \x02(\r\x12\x0b\n\x03url\x18\x03 \x02(\t\x12\x0f\n\x07version\x18\x04 \x03(\t\x12\x10\n\x08proxy_ip\x18\x05 \x01(\t\x12\x12\n\nproxy_port\x18\x06 \x01(\r\x12\r\n\x05is_ts\x18\x07 \x02(\r\x12\x11\n\ttimestamp\x18\x08 \x02(\r\x12 \n\x18monday_0oclock_timestamp\x18\t \x02(\r\x12\x10\n\x08\x63\x64n_list\x18\n \x03(\t')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'p10min_pb_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _globals['_CS_10800']._serialized_start=36
  _globals['_CS_10800']._serialized_end=79
  _globals['_SC_10801']._serialized_start=82
  _globals['_SC_10801']._serialized_end=288
# @@protoc_insertion_point(module_scope)



import sys
import json
import yaml
from shutil import copy
from pathlib import Path
from importlib.resources import files

from .classes import Client, UserConfig, ClientConfig


# package-incuded filepaths
CONFIG_DATA_PATH = files("azlassets").joinpath("config")
YAML_TEMPLATE_PATH = CONFIG_DATA_PATH.joinpath("user_config_template.yml")
CLIENT_CONFIG_PATH = CONFIG_DATA_PATH.joinpath("client_config.json")

# cwd-relative filepaths
YAML_CONFIG_PATH = Path("config") / "user_config.yml"


def load_user_config() -> UserConfig:
	if not YAML_CONFIG_PATH.exists():
		print("Userconfig does not exist. A new one will be created.")
		print("Note that the useragent is empty and it is advised to set one.")
		YAML_CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
		copy(YAML_TEMPLATE_PATH, YAML_CONFIG_PATH)
		return load_user_config()

	with open(YAML_CONFIG_PATH, 'r', encoding='utf8') as file:
		yamlconfig = yaml.safe_load(file)

	try:
		userconfig = UserConfig(
			useragent=yamlconfig['useragent'],
			download_isblacklist=yamlconfig['download-folder-listtype'] == 'blacklist',
			download_filter=yamlconfig['download-folder-list'],
			extract_isblacklist=yamlconfig['extract-folder-listtype'] == 'blacklist',
			extract_filter=yamlconfig['extract-folder-list'],
			asset_directory=yamlconfig['asset-directory'],
			extract_directory=yamlconfig['extract-directory'],
		)
	except KeyError:
		print("There is an error inside the userconfig file. Delete it or change the wrong values.")
		sys.exit(1)

	return userconfig


def load_client_config(client: Client) -> ClientConfig:
	with open(CLIENT_CONFIG_PATH, 'r', encoding='utf8') as f:
		configdata = json.load(f)

	if not client.name in configdata:
		raise NotImplementedError(f'Client {client.name} has not been configured yet.')

	config = configdata[client.name]
	try:
		clientconfig = ClientConfig(config['gateip'], config['gateport'], config['cdnurl'])
	except KeyError:
		print("The clientconfig has been wrongly configured.")
		sys.exit(1)

	return clientconfig



#!/usr/bin/env python
import itertools, json
from argparse import ArgumentParser
from pathlib import Path
import multiprocessing as mp
from typing import Iterable

from azlassets import __version__, imgrecon, config, versioncontrol
from azlassets.classes import Client, VersionType


def get_difflog_versionlist(parent_directory: Path, vtype: VersionType) -> list[str]:
	difflog_dir = Path(parent_directory, "difflog", vtype.name.lower())
	# make sure latest.json does not exist anymore before retrieving file list
	versioncontrol.legacy_rename_latest_difflog(difflog_dir)
	difflog_versionlist = [path.stem for path in difflog_dir.glob("*.json")]
	return difflog_versionlist

def get_diff_files(parent_directory: Path, vtype: VersionType, version_string: str | None = None) -> Iterable[str]:
	if not version_string:
		version_string = versioncontrol.get_latest_versionstring(vtype, parent_directory)

	if version_string:
		difflog_path = Path(parent_directory, "difflog", vtype.name.lower(), version_string+".json")
		if difflog_path.exists():
			with open(difflog_path, "r", encoding="utf8") as f:
				diffdata = json.load(f)
				filtered_success_file_entries = filter((lambda i: i[1] != "Deleted"), diffdata["success_files"].items())
				filenames = [i[0] for i in filtered_success_file_entries]
				return filenames
		elif version_string is not None:
			raise FileExistsError(f"There is no difflog '{version_string}' for version type '{vtype.name}'")
	return []


def restore_painting(image, abpath: Path, imgname: str, do_retry: bool):
	mesh = imgrecon.load_mesh(str(abpath), imgname+'-mesh')
	if mesh is not None:
		return imgrecon.recon(image, mesh)

	if not do_retry:
		return image

	# for some images, the mesh is in the non-tex asset bundle for some reason
	if abpath.name.endswith('_tex'):
		return restore_painting(image, abpath.with_name(abpath.name[:-4]), imgname, False)

	return restore_painting(image, abpath.with_name(abpath.name+'_tex'), imgname, False)

def try_safe_image(image, target: Path) -> Path:
	target.parent.mkdir(parents=True, exist_ok=True)
	while True:
		if target.exists():
			print(f'ERROR: Tried to save "{target}", but the file already exists.')
			target = target.with_name(target.stem + "_" + target.suffix)
		else:
			image.save(target)
			return target

def extract_assetbundle(rootfolder: Path, filepath: str, targetfolder: Path) -> Path | None:
	all_images = []
	abpath = Path(rootfolder, filepath)
	for reader, texture2d in imgrecon.load_images(str(abpath)):
		name = texture2d.m_Name
		if name == 'UISprite': continue # skip the UISprite element
		if 'char' in (reader.container or ''): continue # skip image if its of a chibi

		image = texture2d.image
		if filepath.split('/')[0] == 'painting':
			image = restore_painting(image, abpath, name, True)
		all_images.append((image, name))

	if len(all_images) == 1:
		image, imgname = all_images[0]
		target = Path(Path(targetfolder, filepath).parent, imgname+'.png')
		return try_safe_image(image, target)

	if len(all_images) > 1:
		img_target_dir = Path(Path(targetfolder, filepath).parent, abpath.name)
		for image, imgname in all_images:
			target = Path(img_target_dir, imgname+'.png')
			try_safe_image(image, target)
		return img_target_dir


def extract_by_client(client: Client, target_version: str | None = None, do_iterative_version_check: bool = False):
	userconfig = config.load_user_config()
	client_directory = Path(userconfig.asset_directory, client.name)
	extract_directory = Path(userconfig.extract_directory, client.name)

	downloaded_files_collection = []
	if target_version is None or target_version == "latest":
		target_versiontypes = [VersionType.AZL, VersionType.PAINTING, VersionType.MANGA, VersionType.PIC]
	else:
		target_versiontypes = [VersionType.AZL]

	for vtype in target_versiontypes:
		if do_iterative_version_check:
			version_strings = []
			for vstring in get_difflog_versionlist(client_directory, vtype):
				if versioncontrol.compare_version_string(vstring, target_version) or vstring == target_version:
					version_strings.append(vstring)
		else:
			version_strings = [target_version]

		for vstring in version_strings:
			downloaded_files = get_diff_files(client_directory, vtype, vstring)
			downloaded_files_collection.append(downloaded_files)
	downloaded_files_collection = itertools.chain(*downloaded_files_collection)

	def _filter(assetpath: str) -> bool:
		if assetpath.split('/')[0] in userconfig.extract_filter:
			return (not userconfig.extract_isblacklist)
		return userconfig.extract_isblacklist

	with mp.Pool(processes=mp.cpu_count()-1) as pool:
		for assetpath in filter(_filter, downloaded_files_collection):
			pool.apply_async(extract_assetbundle, (Path(client_directory, 'AssetBundles'), assetpath, extract_directory,))

		# explicitly join pool
		# this causes the pool to wait for all asnyc tasks to complete
		pool.close()
		pool.join()

def extract_single_assetbundle(client: Client, assetpath: str) -> Path | None:
	userconfig = config.load_user_config()
	client_directory = Path(userconfig.asset_directory, client.name, 'AssetBundles')
	extract_directory = Path(userconfig.extract_directory, client.name)
	return extract_assetbundle(client_directory, assetpath, extract_directory)


def main():
	print(f"Running Azurlane file extractor v{__version__}.")

	# setup argument parser
	parser = ArgumentParser(description="Extracts image assets as pngs.",
		epilog="If '-f/--filepath' is not set, all files from the latest update will be extracted.")
	parser.add_argument("client", type=str, choices=Client.__members__,
		help="client to extract files of")
	parser.add_argument("-f", "--filepath", type=str,
		help="Path to the file to extract only this single file")
	parser.add_argument("-v", "--version", type=str,
		help="Extract files of a specific version (Currently only applies to AZL Versiontype!)")
	parser.add_argument("-u", "--until-version", type=str,
		help="Extract files from the latest until a specific version (Currently only applies to AZL Versiontype!)")
	args = parser.parse_args()

	# parse arguments and execute
	client = Client[args.client]
	if filepath := args.filepath:
		extract_single_assetbundle(client, filepath)
	else:
		if version := args.until_version:
			extract_by_client(client, version, True)
		elif version := args.version:
			extract_by_client(client, version)
		else:
			extract_by_client(client)

if __name__ == "__main__":
	main()



import re
from UnityPy import AssetsManager
from UnityPy.enums import ClassIDType
from PIL import Image


VR = re.compile(r'v ')
TR = re.compile(r'vt ')
SR = re.compile(r' ')
def recon(src, mesh):
	sx, sy = src.size
	c = map(SR.split, list(filter(TR.match, mesh))[1::2])
	p = map(SR.split, list(filter(VR.match, mesh))[1::2])
	c = [(round(float(a[1])*sx), round((1-float(a[2]))*sy)) for a in c]
	p = [(-int(float(a[1])), int(float(a[2]))) for a in p]
	my = max(y for x, y in p)
	p = [(x, my-y) for x, y in p[::2]]
	cp = [(l+r, p) for l, r, p in zip(c[::2], c[1::2], p)]
	ox, oy = zip(*[(r-l+p, b-t+q) for (l, t, r, b), (p, q) in cp])
	out = Image.new('RGBA', (max(ox), max(oy)))
	for c, p in cp: out.paste(src.crop(c), p)
	return out

def load_mesh(filepath, require_name=None):
	am = AssetsManager(filepath)
	for obj in am.objects:
		if obj.type == ClassIDType.Mesh:
			objdata = obj.read()
			if require_name and require_name != objdata.m_Name:
				continue
			data = objdata.export().splitlines()
			return data

def load_images(filepath: str):
	am = AssetsManager(filepath)
	for obj in am.objects:
		if obj.type == ClassIDType.Texture2D:
			yield obj, obj.read()
